﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;
using System.IO;

using LibUsbDotNet;
using LibUsbDotNet.Main;
using LibUsbDotNet.DeviceNotify;


namespace SQP_I2Cscan
{
	public partial class MainForm : Form
	{
		// ATB-USBasp device description parameters:
		const string ATBUSBasp_NAME				=	"USBasp";
		const short ATBUSBasp_PID				=	0x05DC;
		const short ATBUSBasp_VID				=	0x16C0;
		const string ATBUSBasp_SERIAL_STRING	=	"www.atnel.pl";

		// I2C devices description file:
		const string I2C_DESCRIPTION_FILE		=	"i2c_table.txt";
		
		// Addresses range to scan:
		const byte I2C_7BIT_ADDRESS_MIN			=	0x08;
		const byte I2C_7BIT_ADDRESS_MAX			=	0x77;
		 	
		// ATB-USBasp LibUSB commands I2C:
		const byte USBASP_I2C_SLA_CHECK			=	100;
		const byte USBASP_FUNC_I2C_BITRATE		=	101;


		struct I2CDeviceDescription
    	{
			public byte Address;
			public string Description;
		};


		private static UsbRegistry MyUsbRegistry = null;
		private static UsbDevice MyUsbDevice = null;
		private static UsbDeviceFinder MyUsbFinder = new UsbDeviceFinder(ATBUSBasp_VID, ATBUSBasp_PID);
		public static IDeviceNotifier UsbDeviceNotifier = null;
		private bool ATBUSBasp = false;
		private bool I2CDescriptionInfo = false;
		private List<I2CDeviceDescription> listNew = new List<I2CDeviceDescription>();



		public MainForm()
		{
			InitializeComponent();
			SplashForm Splash = new SplashForm();
			Splash.ShowDialog();
			progressBarScan.Minimum = I2C_7BIT_ADDRESS_MIN;
			progressBarScan.Maximum = I2C_7BIT_ADDRESS_MAX;			

			ATBUSBasp = FindATBUSBaspDevice();
			SetATBUSBaspStatus();

			UsbDeviceNotifier = DeviceNotifier.OpenDeviceNotifier();
			UsbDeviceNotifier.OnDeviceNotify += OnDeviceNotifyEvent;
			UsbDeviceNotifier.Enabled = true;

			I2CDescriptionInfo = readI2CDeviceDescriptionFile();
		}


		private void OnDeviceNotifyEvent(object sender, DeviceNotifyEventArgs e)
		{
			if ((e.Device.IdProduct == ATBUSBasp_PID) && (e.Device.IdVendor == ATBUSBasp_VID))
			{
				switch (e.EventType)
				{
				case EventType.DeviceArrival:
											if (UsbDevice.IsLinux) ATBUSBasp = true;
											else ATBUSBasp = FindATBUSBaspDevice();
											break;
											
					case EventType.DeviceRemoveComplete:
											ATBUSBasp = false;
											break;
				}
				SetATBUSBaspStatus();
			}
		}


		private bool FindATBUSBaspDevice()
		{
			bool Status = false;
			
			UsbRegDeviceList allDevices = UsbDevice.AllDevices;
            foreach (UsbRegistry usbRegistry in allDevices)
            {
           		try
            	{
           			if ((usbRegistry.Vid == ATBUSBasp_VID) &&
						(usbRegistry.Pid == ATBUSBasp_PID) &&
						((UsbDevice.IsLinux) ||
							((usbRegistry.Name == ATBUSBasp_NAME) &&
							(usbRegistry.Device.Info.SerialString == ATBUSBasp_SERIAL_STRING))))
           			{
		            	MyUsbRegistry = usbRegistry;
		            	Status = true;
           			}
            	}
            	catch {}
			}
            UsbDevice.Exit();
            return Status;
		}


		private bool CheckATBUSBaspDevice()
		{
			return MyUsbRegistry.IsAlive;
		}


		private bool OpenATBUSBaspDevice()
		{
			bool status = false;
			try
			{
				if (UsbDevice.IsLinux)
				{
					if ((MyUsbDevice = UsbDevice.OpenUsbDevice(MyUsbFinder)) != null) status = true;
				}
				else
				{
					if (MyUsbRegistry.Open(out MyUsbDevice)) status = true;
				}
    		}
	       	catch {}

    	    return status;
		}


		private void CloseATBUSBaspDevice()
		{
			try
			{
				if (MyUsbDevice.IsOpen) MyUsbDevice.Close();
			}
			catch {}
		}


		private bool SendToATBUSBaspDevice ( UsbSetupPacket setupPacket, byte[] dataBuffer )
		{
			int Tranferred;
			
			if (MyUsbDevice.ControlTransfer(ref setupPacket, dataBuffer, 254, out Tranferred))
			{
			 	if (Tranferred < 5)
			 	{
			 		return true;
			 	}
			}
			return false;
		}

		
		private bool i2cSetSCL ( short clock )
		{
			byte[] dataBuffer = new byte[254];
			UsbSetupPacket setupPacket = new UsbSetupPacket((byte)(UsbCtrlFlags.Direction_In | UsbCtrlFlags.RequestType_Vendor),
															USBASP_FUNC_I2C_BITRATE,
															clock,
															0,
															254);
			
			return SendToATBUSBaspDevice(setupPacket, dataBuffer);
		}
		
		
		private bool i2cCheck ( byte address )
		{
			byte[] dataBuffer = new byte[254];
			UsbSetupPacket setupPacket = new UsbSetupPacket((byte)(UsbCtrlFlags.Direction_In | UsbCtrlFlags.RequestType_Vendor),
			                                                USBASP_I2C_SLA_CHECK,
			                                                address,
			                                                0,
			                                                254);
			
			return SendToATBUSBaspDevice(setupPacket, dataBuffer);
		}


		void SetATBUSBaspStatus()
		{
			if (ATBUSBasp)
			{
				buttonScan.Enabled = true;
				toolStripStatusLabelATBUSBaspStatus.BackColor = Color.YellowGreen;
				toolStripStatusLabelATBUSBaspStatus.Text = "ATB-USBasp Connected";
			}
			else
			{
				buttonScan.Enabled = false;
				toolStripStatusLabelATBUSBaspStatus.BackColor = Color.Red;
				toolStripStatusLabelATBUSBaspStatus.Text = "ATB-USBasp Disconnected";
			}
		}


		private string ChceckI2CDeviceDescription ( byte address )
		{
			string description = "";

			if (I2CDescriptionInfo)
			{
				I2CDeviceDescription device = listNew.Find(x => x.Address == address);
				description = device.Description;
			}
			return description;
		}


		private bool readI2CDeviceDescriptionFile()
		{
			bool status = false;
			string line;
			I2CDeviceDescription device;

			try
			{
				using (StreamReader reader = new StreamReader(I2C_DESCRIPTION_FILE))
				{
					while ((line = reader.ReadLine()) != null)
					{
						try
						{
							device.Address = Convert.ToByte(line.Substring(0, 2), 16);
							device.Description = line.Substring(3);
							listNew.Add(device);
							status = true;
						}
						catch {}
					}
				}
			}
			catch {}
			return status;
		}


		private void ShowClosingInfo()
		{
			while (statusStrip.Items.Count > 1)	statusStrip.Items.RemoveAt(1);
			statusStrip.Items [0].Text = "Closing SQP-I2Cscan application ...";
			statusStrip.Items [0].BackColor = Color.DimGray;
			statusStrip.Items [0].ForeColor = Color.White;
			statusStrip.Items [0].Size = new Size (636, 32);
			statusStrip.Refresh ();
		}


		// Sets the progress bar value, without using Windows Aero animation
		private void SetProgressNoAnimation(ProgressBar pb, int value)
		{   
			// To get around this animation, we need to move the progress bar backwards.
			if (value == pb.Maximum)
			{												// Special case (can't set value > Maximum).
				pb.Value = value;							// Set the value
				pb.Value = value - 1;						// Move it backwards
			}
			else
			{
				pb.Value = value + 1;						// Move past
			}
        	pb.Value = value;								// Move to correct value
		}


		void ButtonScanClick ( object sender, EventArgs e )
		{
			buttonScan.Enabled = false;
			trackBarI2CSpeed.Enabled = false;
			listViewI2CDevices.Items.Clear();
			if (OpenATBUSBaspDevice())
			{
				if (i2cSetSCL((short)trackBarI2CSpeed.Value))
				{
					progressBarScan.Value = I2C_7BIT_ADDRESS_MIN;
					progressBarScan.Visible = true;
					labelProgress.Text = "0x" + I2C_7BIT_ADDRESS_MIN.ToString("X2");
					labelProgress.Visible = true;
					for (byte i = I2C_7BIT_ADDRESS_MIN; i <= I2C_7BIT_ADDRESS_MAX; i++)
					{
						SetProgressNoAnimation(progressBarScan, i);
						if (i2cCheck((byte)(i << 1)))
						{
							int pos = listViewI2CDevices.Items.Count;
							listViewI2CDevices.Items.Add((pos + 1).ToString());
							listViewI2CDevices.Items[pos].SubItems.Add("0x" + i.ToString("X2") + " (" + i.ToString() + ")");
							listViewI2CDevices.Items[pos].SubItems.Add("0x" + (i << 1).ToString("X2") + " (" + (i << 1).ToString() + ")");
							listViewI2CDevices.Items[pos].SubItems.Add(ChceckI2CDeviceDescription(i));
						}
						labelProgress.Text = "0x" + i.ToString("X2");
						Application.DoEvents();
					}
					listViewI2CDevices.Columns[3].Width = -2;
					labelProgress.Visible = false;
					progressBarScan.Visible = false;
					if (listViewI2CDevices.Items.Count == 0) MessageBox.Show("No I2C device found.", "SQP-I2Cscan", MessageBoxButtons.OK, MessageBoxIcon.Information);
					else MessageBox.Show("Found I2C devices: " + listViewI2CDevices.Items.Count.ToString() + ".", "SQP-I2Cscan", MessageBoxButtons.OK, MessageBoxIcon.Information);
				}
				else
				{
					MessageBox.Show("Can't set SCK Clock.", "SQP-I2Cscan", MessageBoxButtons.OK, MessageBoxIcon.Error);
				}
				CloseATBUSBaspDevice();
			}
			else
			{
				MessageBox.Show("Can't open ATB-USBasp device.", "SQP-I2Cscan", MessageBoxButtons.OK, MessageBoxIcon.Error);
			}
			trackBarI2CSpeed.Enabled = true;
			buttonScan.Enabled = true;
		}


		void TrackBarI2CSpeedValueChanged ( object sender, EventArgs e )
		{
			labelSCLSpeed.Text = trackBarI2CSpeed.Value.ToString();	
		}


		void MainFormFormClosing ( object sender, FormClosingEventArgs e )
		{
			ShowClosingInfo ();
			UsbDeviceNotifier.Enabled = false;
			UsbDeviceNotifier.OnDeviceNotify -= OnDeviceNotifyEvent;
			try
			{
				if (MyUsbDevice.IsOpen) MyUsbDevice.Close();
			}
			catch {}
//			Application.ExitThread();
			Environment.Exit(0);
//			if (System.Windows.Forms.Application.MessageLoop) System.Windows.Forms.Application.Exit();
//			else System.Environment.Exit(0);
		}
	}
}
